/***************************************************
 * CS182, Assignment 6: Model Merging
 * ModelMerge.java
 * Author: J.G. Makin
 * Adapted by Leon Barrett
 * Last updated: 4/02/07
 **************************************************/



import java.util.*;

/**
 * The children of an AffixTreeNode are stored in a hashtable. Remember that
 * things one gets out of a Hashtable need to be cast to whatever type they are!
 * 
 * In addition it has affix-specific fields and rules that have up through this
 * node as an affix.
 */
class AffixTreeNode {
	/** The map from next affix symbol to children. */
    HashMap<String,AffixTreeNode> children = new HashMap<String,AffixTreeNode>();
    /** Number of steps from root; root depth = 0. */
    int depth = 0;
    /** The parent of this node.  Is null iff this is the root. */
    AffixTreeNode parent;
    /** The most recent word of the affix. */
    String affixWord;
    /** All rules that find this node on their path. */
    HashSet<Rule> rules = new HashSet<Rule>();
    /** All rules that end in this node. */
    HashSet<Rule> terminatingRules = new HashSet<Rule>();
    /** True iff this node is in a prefix tree. */
    boolean isPrefix;
    
    /** Create a new node with this most recent affix word and parent.
     * 
     * @param affixWord The latest word in the affix.
     * @param parent
     */
    AffixTreeNode(String affixWord, AffixTreeNode parent, boolean isPrefix){
        this.affixWord = affixWord;
        this.parent = parent;
        if (parent!=null)
        	depth = parent.depth + 1;
        this.isPrefix = isPrefix;
    }
    
    
    void addChild(AffixTreeNode c){
        children.put(c.affixWord, c);
    }
    
    
    AffixTreeNode getChild(String key){
        return (AffixTreeNode)children.get(key);
    }
    
    
    void removeChild(String key){
        children.remove(key);
    }
    
    
    Collection<AffixTreeNode> getChildren(){
        return children.values();
    }
    
    
    /** Get the most recent word in this node's affix. */
    String getAffixWord(){
        return(affixWord);
    }
    
    
    /** Get all rules that include this affix. */
    HashSet<Rule> getRules(){
        return rules;
    }
    
    
    boolean hasMultipleRules(){
        int numberOfRules = rules.size();
        return (numberOfRules > 1);    
    }
    
    
    /** Non-recursively add a rule to the list of rules that pass through this node. */
    void addRule(Rule r){
        // We add the whole rule to the vector so that, when we pass
        //  a node to a Merge, the latter has access to r.numTimesUsed.
        this.rules.add(r);
        if (r.prefixPath().size()==depth)
        	this.terminatingRules.add(r);
    }
    
    
    /** Non-recursively remove a rule from the list of rules that pass through this node. */
    void removeRule(Rule r){
        this.rules.remove(r);
        if (r.prefixPath().size()==depth)
        	this.terminatingRules.remove(r);
    }
    
    
    /** Length of affix */
    int affixLength() {
    	return depth - 1;
    }
    
    
    /** Path to this node from the root. */
    ArrayList<String> rootPath() {
    	ArrayList<String> rp;
    	if (parent!=null) {
    		rp = parent.rootPath();
        	rp.add(affixWord);
    	}
    	else
    		rp = new ArrayList<String>();
    	return rp;
    }
    
    
    /** Get the complete affix for this node. If this is a prefix node,
     * then it includes the LHS of the rule as the first element. */
    ArrayList<String> fullAffix() {
    	ArrayList<String> rp = rootPath();
    	// reverse the order so we put the suffix in the right order
    	if (!isPrefix) {
	    	rp.remove(0);
	    	Collections.reverse(rp);
    	}
    	return rp;
    }
    
    
    public String toString() {
    	ArrayList<String> affix = fullAffix();
    	String str = "... ";
    	if (isPrefix) {
    		str = affix.get(0) + " -> ";
    		affix.remove(0);
    	}
    	for (String s : affix) {
    		str += s + " ";
    	}
    	str = str.substring(0, str.length()-1);
    	if (isPrefix)
    		str += " ...";
    	return str;
    }
}